"""Hatchling plugin to create optional-dependencies pinned to minimum versions."""

from __future__ import annotations

import os
from importlib import metadata
from pathlib import Path

import tomlkit
from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl

from .util import sub_min_compatible_version

try:
    __version__ = metadata.version("hatch-min-requirements")
except metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"
__author__ = "Talley Lambert"

__all__ = ["MinRequirementsMetadataHook", "patch_pyproject", "sub_min_compatible_version"]


MIN_REQS_EXTRA = os.getenv("MIN_REQS_EXTRA_NAME", "min-reqs")


class MinRequirementsMetadataHook(MetadataHookInterface):
    """Hatchling metadata hook to populate optional-dependencies from dependencies."""

    PLUGIN_NAME = "min_requirements"

    def update(self, metadata: dict) -> None:
        """Update the project.optional-dependencies metadata."""
        min_reqs = [sub_min_compatible_version(dep) for dep in metadata["dependencies"]]
        metadata["optional-dependencies"][MIN_REQS_EXTRA] = min_reqs


@hookimpl  # type: ignore
def hatch_register_metadata_hook() -> type[MetadataHookInterface]:
    return MinRequirementsMetadataHook


def patch_pyproject(
    path: str | Path = "pyproject.toml", extra_name: str = MIN_REQS_EXTRA, tab: str = " " * 4
) -> None:
    """Directly modify pyproject.toml to add min-reqs optional-dependency.

    The pyproject.toml file is modified in place, with a backup created as
    "pyproject.toml.BAK".

    Parameters
    ----------
    path : str or Path
        Path to the pyproject.toml file.
    extra_name : str, optional
        Name of the extra to add to the optional-dependencies table, by default
        'min-reqs'. Can be set with the MIN_REQS_EXTRA_NAME environment variable.
    tab : str, optional
        String to use for indentation, by default four spaces.
    """
    with open(path) as f:
        pyproject_text = f.read()

    doc = tomlkit.parse(pyproject_text)

    # get the dependencies and create min-reqs
    project = doc.get("project", {})
    deps = project.get("dependencies", [])
    if not deps:
        return
    min_reqs = [sub_min_compatible_version(dep) for dep in deps]

    # Ensure optional-dependencies subsection exists
    if "optional-dependencies" not in project:
        project["optional-dependencies"] = tomlkit.table()

    # Add the min-reqs entry
    project["optional-dependencies"][extra_name] = min_reqs

    # backup original pyproject.toml
    with open(Path(path).with_suffix(".toml.BAK"), "w") as f:
        f.write(pyproject_text)

    with open(path, "w") as f:
        f.write(tomlkit.dumps(doc))



import pytest

from hatch_min_requirements.util import sub_min_compatible_version
from hatch_min_requirements import _fetch

MIN_NUMPY = "1.3.0"
# these may change if the min versions change upstream
PARAMS = [
    ("numpy", f"numpy=={MIN_NUMPY}", "numpy"),
    ("numpy!=1.3.6", f"numpy=={MIN_NUMPY}", "numpy!=1.3.6"),
    ("numpy!=1.3.0", "numpy==1.4.1", "numpy!=1.3.0"),
    ("numpy~=1.7", "numpy==1.7.0", "numpy==1.7"),
    ("numpy===1.7", "numpy==1.7.0", "numpy==1.7"),  # current behavior, perhaps weird
    ("numpy>=1.5.0", "numpy==1.5.0", "numpy==1.5.0"),
    ("numpy >=1.5.0", "numpy ==1.5.0", "numpy ==1.5.0"),
    ("numpy[extra]", f"numpy[extra]=={MIN_NUMPY}", "numpy[extra]"),
    ("numpy>1.3", "numpy==1.4.1", "numpy>1.3"),
    ("numpy[extra]>1.3", "numpy[extra]==1.4.1", "numpy[extra]>1.3"),
    (
        "numpy[extra]>1.3; python_version=='3.7'",
        "numpy[extra]==1.4.1; python_version=='3.7'",
        "numpy[extra]>1.3; python_version=='3.7'",
    ),
]


@pytest.mark.parametrize("offline", [True, False])
@pytest.mark.parametrize("package, online_expectation, offline_expectation", PARAMS)
def test_sub_min_compatible_version(
    package: str, online_expectation: str, offline_expectation: str, offline: bool
) -> None:
    expectation = offline_expectation if offline else online_expectation
    assert sub_min_compatible_version(package, offline=offline) == expectation


def test_pin_unconstrained() -> None:
    assert (
        sub_min_compatible_version("numpy", offline=False, pin_unconstrained=True)
        == f"numpy=={MIN_NUMPY}"
    )
    assert (
        sub_min_compatible_version("numpy", offline=False, pin_unconstrained=False) == "numpy"
    )
    assert (
        sub_min_compatible_version("numpy", offline=True, pin_unconstrained=False) == "numpy"
    )
    assert (
        sub_min_compatible_version("numpy", offline=True, pin_unconstrained=True) == "numpy"
    )


@pytest.mark.parametrize("use_pip", [True, False])
def test_fetch_versions(use_pip: bool, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(_fetch, "TRY_PIP", use_pip)

    _fetch.fetch_available_versions.cache_clear()
    available = _fetch.fetch_available_versions("numpy")
    assert available[-1] == MIN_NUMPY
    assert available[-4:] == ["1.5.1", "1.5.0", "1.4.1", "1.3.0"]



from __future__ import annotations

import os
import re
import urllib.request
from contextlib import suppress
from functools import lru_cache
import warnings
from packaging.version import parse as parse_version
from packaging.version import Version as PackagingVersion

TRY_PIP = os.getenv("MIN_REQS_TRY_PIP", "1") in {"1", "true", "True", "yes", "Yes"}

__all__ = ["fetch_available_versions"]


@lru_cache
def fetch_available_versions(package_name: str) -> list[str]:
    if TRY_PIP:
        with suppress(ImportError):
            import pip  # noqa: F401

            return fetch_available_versions_pip(package_name)

    return fetch_available_versions_std_lib(package_name)


@lru_cache
def fetch_available_versions_pip(package_name: str) -> list[str]:
    """Return a list of available versions for the given `package_name`.

    This uses `pip index versions` to get the available versions, which requires
    pip >= 21.2.

    Parameters
    ----------
    package_name : str
        The name of the package.

    Returns
    -------
    list[str]
        A list of available version strings for the package.
    """
    import subprocess

    try:
        output = subprocess.check_output(
            ["pip", "index", "versions", package_name],
            text=True,
            stderr=subprocess.PIPE,
        )
        avail = output.split("Available versions:")[1].split("\n")[0].strip()
        return avail.split(", ")
    except subprocess.CalledProcessError as e:  # pragma: no cover
        warnings.warn(
            f"Failed to fetch available versions for {package_name}: {e}",
            RuntimeWarning,
            stacklevel=2,
        )
        return []


@lru_cache
def fetch_available_versions_std_lib(package_name: str) -> list[str]:
    """Fetch all versions of a package from the simple PyPI index."""
    url = f"https://pypi.org/simple/{package_name}/"
    try:
        with urllib.request.urlopen(url) as response:
            html = response.read().decode("utf-8")
            versions: set[PackagingVersion] = set()
            for link in re.findall(r">([^>]+)</a>", html):
                with suppress(ValueError):
                    version = parse_version(parse_filename(link)[1])
                    if not version.is_prerelease:
                        versions.add(version)
            return [str(x) for x in sorted(versions, reverse=True)]
    except Exception as e:  # pragma: no cover
        warnings.warn(
            f"Failed to fetch available versions for {package_name}: {e}",
            RuntimeWarning,
            stacklevel=2,
        )
        return []


# ---------- Below vendored from pypi-simple ---------------------
#
# https://github.com/jwodder/pypi-simple
#
# The MIT License (MIT)
#
# Copyright (c) 2018-2024 John Thorvald Wodder II and contributors

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


PROJECT_NAME = r"[A-Za-z0-9](?:[A-Za-z0-9._-]*[A-Za-z0-9])?"
PROJECT_NAME_NODASH = r"[A-Za-z0-9](?:[A-Za-z0-9._]*[A-Za-z0-9])?"
VERSION = r"[A-Za-z0-9_.!+-]+?"
VERSION_NODASH = r"[A-Za-z0-9_.!+]+?"
ARCHIVE_EXT = r"\.(?:tar|tar\.(?:bz2|gz|lz|lzma|xz|Z)|tbz|tgz|tlz|txz|zip)"
PLAT_NAME = r"(?:aix|cygwin|darwin|linux|macosx|solaris|sunos|[wW]in)[-.A-Za-z0-9_]*"
PYVER = r"py[0-9]+\.[0-9]+"

#: Regexes for package filenames that can be parsed unambiguously
GOOD_PACKAGE_RGXN = [
    # See <https://setuptools.readthedocs.io/en/latest
    #      /formats.html#filename-embedded-metadata>:
    (
        "egg",
        re.compile(
            r"^(?P<project>{})-(?P<version>{})(?:-{}(?:-{})?)?\.egg$".format(
                PROJECT_NAME_NODASH, VERSION_NODASH, PYVER, PLAT_NAME
            )
        ),
    ),
    # See <http://ftp.rpm.org/max-rpm/ch-rpm-file-format.html>:
    # (The architecture pattern is mainly just a guess based on what's
    # currently on PyPI.)
    (
        "rpm",
        re.compile(
            r"^(?P<project>{})-(?P<version>{})-[^-]+\.[A-Za-z0-9._]+\.rpm$".format(
                PROJECT_NAME, VERSION_NODASH
            )
        ),
    ),
    # Regex adapted from <https://github.com/pypa/pip/blob/18.0/src/pip/_internal/wheel.py#L569>:
    (
        "wheel",
        re.compile(
            r"^(?P<project>{})-(?P<version>{})(-[0-9][^-]*?)?" r"-.+?-.+?-.+?\.whl$".format(
                PROJECT_NAME_NODASH, VERSION_NODASH
            )
        ),
    ),
]

#: Partial regexes for package filenames with ambiguous grammars.  If a hint as
#: to the expected project name is given, it will be prepended to the regexes
#: when trying to determine a match; otherwise, a generic pattern that matches
#: all project names will be prepended.
BAD_PACKAGE_BASES = [
    # See <https://github.com/python/cpython/blob/v3.7.0/Lib/distutils/command/bdist_dumb.py#L93>:
    (
        "dumb",
        re.compile(rf"-(?P<version>{VERSION})\.{PLAT_NAME}{ARCHIVE_EXT}$"),
    ),
    # See <https://github.com/python/cpython/blob/v3.7.0/Lib/distutils/command/bdist_msi.py#L733>:
    (
        "msi",
        re.compile(rf"-(?P<version>{VERSION})\.{PLAT_NAME}(?:-{PYVER})?\.msi$"),
    ),
    ("sdist", re.compile(rf"-(?P<version>{VERSION}){ARCHIVE_EXT}$")),
    # See <https://github.com/python/cpython/blob/v3.7.0/Lib/distutils/command/bdist_wininst.py#L292>:
    (
        "wininst",
        re.compile(rf"-(?P<version>{VERSION})\.{PLAT_NAME}(?:-{PYVER})?\.exe$"),
    ),
]

#: Regexes for package filenames with ambiguous grammars, using a generic
#: pattern that matches all project names
BAD_PACKAGE_RGXN = [
    (pkg_type, re.compile("^(?P<project>" + PROJECT_NAME + ")" + rgx.pattern))
    for pkg_type, rgx in BAD_PACKAGE_BASES
]


def parse_filename(filename: str, project_hint: str | None = None) -> tuple[str, str, str]:
    for pkg_type, rgx in GOOD_PACKAGE_RGXN:
        m = rgx.match(filename)
        if m:
            return (m.group("project"), m.group("version"), pkg_type)
    if project_hint is not None:  # pragma: no cover
        proj_rgx = re.sub(r"[^A-Za-z0-9]+", "[-_.]+", project_hint)
        proj_rgx = re.sub(
            r"([A-Za-z])",
            lambda m: "[" + m.group(1).upper() + m.group(1).lower() + "]",
            proj_rgx,
        )
        m = re.match(proj_rgx + r"(?=-)", filename)
        if m:
            project = m.group(0)
            rest_of_name = filename[m.end(0) :]
            for pkg_type, rgx in BAD_PACKAGE_BASES:
                m = rgx.match(rest_of_name)
                if m:
                    return (project, m.group("version"), pkg_type)
    for pkg_type, rgx in BAD_PACKAGE_RGXN:
        m = rgx.match(filename)
        if m:
            return (m.group("project"), m.group("version"), pkg_type)
    raise ValueError(f"Could not parse filename: {filename!r}")  # pragma: no cover



from pathlib import Path

from hatch_min_requirements import patch_pyproject

PYPROJECT = """\
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "myproject"
version = "0.1.0"
description = "My project"
dependencies = ["requests>=2.0.0", "numpy"]
"""

MIN_REQS = """
[project.optional-dependencies]
min-reqs = ["requests==2.0.0", "numpy==1.3.0"]
"""


def test_patch_pyproject(tmp_path: Path) -> None:
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml_BAK = tmp_path / "pyproject.toml.BAK"
    pyproject_toml.write_text(PYPROJECT)

    patch_pyproject(pyproject_toml)
    assert pyproject_toml_BAK.exists()
    assert pyproject_toml_BAK.read_text() == PYPROJECT
    assert pyproject_toml.read_text() == PYPROJECT + MIN_REQS



"""Functions for determining the minimum compatible constrained version of a package."""

from __future__ import annotations

import operator
import os
import re
from contextlib import suppress
import warnings
from typing import TYPE_CHECKING, Callable

from packaging.version import Version as PackagingVersion
from packaging.version import parse as parse_version

from ._fetch import fetch_available_versions

if TYPE_CHECKING:
    from typing import Literal, Sequence, TypeAlias

    # version_cmp   = wsp* <'<=' | '<' | '!=' | '==' | '>=' | '>' | '~=' | '==='>
    VersionCmp: TypeAlias = Literal["<=", "<", "!=", "==", ">=", ">", "~=", "==="]
    # version       = wsp* <( letterOrDigit | '-' | '_' | '.' | '*' | '+' | '!' )+>
    VersionStr: TypeAlias = str
    # version_one   = version_cmp:op version:v wsp* -> (op, v)
    VersionConstraint = tuple[VersionCmp, VersionStr]
    VersionOp = Callable[[PackagingVersion, PackagingVersion], bool]


__all__ = ["sub_min_compatible_version"]


# >>> version_specifier_pattern.findall('~= 0.9, >= 1.0, ===asdfds, <    2.0')
# [('~=', '0.9'), ('>=', '1.0'), ('===', 'asdfds'), ('<', '2.0')]
version_specifier_pattern = re.compile(
    r"""
    (~=|===?|!=|<=?|>=?)  # Comparison operator, captured in a group
    \s*                   # Optional whitespace
    ([^,]*?)              # version string: Non-greedy match for any non-comma character
    (?=                   # Positive lookahead for comma or end of string
        ,\s*              # Comma followed by optional whitespace
        |                 # OR
        $                 # End of string
    )
    """,
    re.VERBOSE,
)


def version_is_compatible(a: PackagingVersion, b: PackagingVersion) -> bool:
    """Return True if package version `a` ~= `b`.

    https://peps.python.org/pep-0440/#compatible-release

    Where ~= is the compatible release operator:
     - ~= x.y        <==>   >= X.Y, == X.*
     - ~= x.y.z      <==>   >= X.Y.Z, == X.Y.*
     - ~= 2.2.post3  <==>   >= 2.2.post3, == 2.*
     - ~= 1.4.5.0    <==>   >= 1.4.5.0, == 1.4.5.*

    """
    # Check if the epoch values match, since different epochs are always incompatible
    if a.epoch != b.epoch:
        return False

    # The ~= operator MUST NOT be used with a single segment version number such as ~=1.
    if len(b.release) < 2:
        raise ValueError(
            "The version for comparison must have at least two segments "
            "(e.g., 1.0, not just 1)"
        )

    # Build the minimum release tuple for compatibility (ignoring pre/post/dev segments)
    # e.g., (1, 2, 0) for 1.2, (1, 2, 3) for 1.2.3
    min_compat_release = b.release[:2] + (0,) * (len(a.release) - 2)

    # Check if `x` is greater than or equal to the minimum compatible version
    # and if `x` starts with the same prefix as `v` for the significant segments
    is_greater_or_equal = a.release >= min_compat_release
    is_same_prefix = a.release[: len(b.release) - 1] == b.release[: len(b.release) - 1]

    return bool(is_greater_or_equal and is_same_prefix)


# Mapping of pep440 version comparison operators to their corresponding functions
# https://peps.python.org/pep-0440/#version-specifiers
COMP_OPS: dict[str, Callable[[PackagingVersion, PackagingVersion], bool]] = {
    "<=": operator.le,
    "<": operator.lt,
    "!=": operator.ne,
    "==": operator.eq,
    ">=": operator.ge,
    ">": operator.gt,
    "~=": version_is_compatible,
    "===": operator.eq,
}


def _min_compatible_version(
    available_versions: Sequence[PackagingVersion],
    constraints: Sequence[VersionConstraint] | None = None,
) -> PackagingVersion:
    """Return the minimum version that satisfies the provided constraints.

    If no constraints are provided, the minimum version is the lowest available version.

    Parameters
    ----------
    available_versions : Sequence[PackagingVersion]
        A sequence of `packaging.version.Version` objects.
    constraints : Sequence[VersionConstraint], optional
        A sequence of version constraint tuples to satisfy, e.g. `[(">=", "1.20")]`.

    Returns
    -------
    PackagingVersion
        The minimum version that satisfies the constraints.

    Examples
    --------
    >>> min_compatible_version([Version("1.3.0"), Version("1.20.1")])
    <Version('1.3.0')>
    >>> min_compatible_version([Version("1.3.0"), Version("1.20.1")], [('>=', '1.5.0')])
    <Version('1.20.1')>

    """
    if not constraints:
        return min(available_versions)

    constraint_ops: list[tuple[VersionOp, PackagingVersion]] = [
        (COMP_OPS[op], parse_version(v)) for op, v in constraints
    ]

    def satisfies_all_constraints(x: PackagingVersion) -> bool:
        return all(op(x, v) for op, v in constraint_ops)

    try:
        min_ver = min(x for x in available_versions if satisfies_all_constraints(x))
    except ValueError:
        raise ValueError(
            f"No available version satisfies the constraints: {constraints}"
        ) from None
    return min_ver


def fetch_min_compatible_version(
    package: str, constraints: Sequence[VersionConstraint] | None = None
) -> str:
    """Fetch minimum version of `package` that satisfies `constraints`.

    This requires network access to fetch the available versions.

    Parameters
    ----------
    package : str
        The name of the package.
    constraints : Sequence[VersionConstraint], optional
        A sequence of version constraints to satisfy, e.g. `[(">=", "1.20")]`.
        By default, `None`.

    Returns
    -------
    str
        The minimum version of the package that satisfies the constraints.

    Examples
    --------
    >>> fetch_min_compatible_version('numpy')
    '1.3.0'
    >>> fetch_min_compatible_version("numpy", [(">=", "1.20")])
    '1.20.0'
    >>> fetch_min_compatible_version("numpy", [(">", "1.20")])
    '1.20.1'
    >>> fetch_min_compatible_version("numpy", [("<", "1.20")])
    '1.3.0'
    >>> fetch_min_compatible_version("numpy", [("<", "1.20"), ("!=", "1.3")])
    '1.4.1'
    """
    available: list[PackagingVersion] = []
    for v in fetch_available_versions(package):
        with suppress(ValueError):  # TODO: warn if we can't parse?
            available.append(parse_version(v))
    if not available:  # pragma: no cover
        raise ValueError(f"No available versions found for package {package!r}")
    return str(_min_compatible_version(available, constraints))


def min_compatible_version_offline(
    constraints: Sequence[VersionConstraint],
) -> str | None:
    """Return the minimum version that satisfies `constraints` without the internet.

    If we don't have internet access, we can't fetch the available versions, and so
    we have to make some assumptions about the available versions in cases with
    non-inclusive lower bounds.
    """
    if not constraints:
        return None

    min_version: PackagingVersion | None = None
    for op, version in constraints:
        # if we find a hard equality constraint, we can return it immediately
        if op in {"==", "==="}:
            return version
        # if we find an inclusive lower bound, check if it's lower than the current min
        if op in {">=", "~="}:
            try:
                v = parse_version(version)
            except ValueError as e:  # pragma: no cover
                warnings.warn(
                    f"Failed to parse version {version!r}: {e}. Skipping.",
                    RuntimeWarning,
                    stacklevel=2,
                )
                continue
            # if we haven't seen any lower bounds yet, or if this lower bound is greater
            # than the current min, update the min
            if min_version is None or v > min_version:
                min_version = v

        # the remaining possible ops are {"<", "<=", ">", "!="}:
        # with upper-bounds, non-inclusive lower bounds, and exclusions,
        # we can't make any assumptions about the minimum version

    return None if min_version is None else str(min_version)


# ###########################################################################
# This is the primary function in this module
# ###########################################################################

YES_LIKE = {"1", "true", "True", "yes", "Yes"}
OFFLINE = os.getenv("MIN_REQS_OFFLINE") in YES_LIKE
PIN_UNCONSTRAINED = os.getenv("MIN_REQS_PIN_UNCONSTRAINED", "1") in YES_LIKE


def sub_min_compatible_version(
    spec: str, offline: bool = OFFLINE, pin_unconstrained: bool = PIN_UNCONSTRAINED
) -> str:
    """Replace the version in a dependency specifier with the min compatible version.

    Parameters
    ----------
    spec : str
        A PEP 440 compliant dependency specifier. Such as:
        - "package-name"
        - "package-name>=1.20"
        - "package-name[extra1,extra2]>=1.20,<2.0"
        - "package-name[extra1,extra2]>=1.20,<2.0 ; python_version == '3.6'"
    offline : bool, optional
        If True, the function will not fetch available versions from the internet.
        By default, False.  Can be overridden by the environment variable
        `MIN_REQS_OFFLINE`.
    pin_unconstrained : bool, optional
        If True, the function will replace unconstrained versions with the min
        compatible version (provided that offline is also `False`). If False, the
        original spec will be returned for unconstrained versions. By default, True.
        Can be overridden by the environment variable `MIN_REQS_PIN_UNCONSTRAINED`.

    Returns
    -------
    str
        A PEP 440 compliant dependency specifier pinned to the minimum
        version that satisfies the constraints in the original specifier.

    Examples
    --------
    These examples would require internet access to fetch the available versions.

    >>> sub_min_compatible_version("numpy")
    'numpy==1.3.0'
    >>> sub_min_compatible_version("numpy>1.3")
    'numpy==1.4.1'
    >>> sub_min_compatible_version("numpy[extra1,extra2]>=1.20,<2.0")
    'numpy[extra1,extra2]==1.20.0'
    >>> sub_min_compatible_version("numpy[extra]<2; python_version == '3.6'")
    "numpy[extra]==1.3.0 ; python_version == '3.6'"
    """
    # Break on semi-colon to separate markers from the name and version spec
    name_spec, *markers = spec.partition(";")

    # Find all version constraints in the first part
    constraints: list[VersionConstraint] = []
    str_idx = None  # index of the first version specifier in the name_spec
    for match in version_specifier_pattern.finditer(name_spec):
        if not str_idx:  # Save the index of the first version specifier
            str_idx = match.start()
        constraints.append(match.groups())  # type: ignore

    # if there are no constraints and we're not pinning unconstrained versions
    # return the original spec as is
    if not constraints and not pin_unconstrained:
        return spec

    # split the name from the extras and version spec
    name_and_extra = name_spec[:str_idx]

    if offline:
        if (min_compat := min_compatible_version_offline(constraints)) is None:
            # if we can't determine a minimum version, just return the original spec
            return spec
        min_compat = min_compat
    else:
        # solve the min version compatible with the constraints, uses the internet
        name = name_and_extra.split("[", 1)[0].strip()
        min_compat = fetch_min_compatible_version(name, constraints)

    return f"{name_and_extra}=={min_compat}{''.join(markers)}"
